#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          09-07-2018
//      Created for Project :   FLAME
"""

from json import dumps
import pytest
from pyramid import testing
from clmcservice.graphapi.views import GraphAPI
from pyramid.httpexceptions import HTTPBadRequest, HTTPNotFound


graph_1_id = None
graph_2_id = None


class TestGraphAPI(object):
    """
    A pytest-implementation test for the Graph API endpoints.
    """

    @pytest.fixture(autouse=True)
    def app_config(self):
        """
        A fixture to implement setUp/tearDown functionality for all tests by initializing configuration structure for the web service and db connection
        """

        self.registry = testing.setUp()
        self.registry.add_settings({"neo4j_host": "localhost", "neo4j_password": "admin", "influx_host": "localhost", "influx_port": 8086})

        yield

        testing.tearDown()

    def test_build(self, db_testing_data):
        """
        Tests the graph build API endpoint

        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data, test database name and the graph db client
        """

        global graph_1_id, graph_2_id

        from_timestamp, to_timestamp, test_db_name, graph_db = db_testing_data

        dc_nodes = set([node["name"] for node in graph_db.nodes.match("ComputeNode")])
        assert dc_nodes == set("DC" + str(i) for i in range(1, 7)), "Compute nodes must have been created by the db_testing_data fixture"

        request = testing.DummyRequest()
        request.body = request.body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of missing request body."

        body = dumps(dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="sfc_i"))
        request = testing.DummyRequest()
        request.params["from"] = 12341412
        request.params["to"] = 12341412
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of invalid request body."

        body = dumps(dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="sfc_1", service_functions="{invalid_json}"))
        request = testing.DummyRequest()
        request.params["from"] = 12341412
        request.params["to"] = 12341412
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of invalid request body."

        service_functions = ["nginx", "minio"]
        body = dumps(dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="sfc_1", service_functions=service_functions))
        request = testing.DummyRequest()
        request.params["from"] = 12341412
        request.params["to"] = 12341412
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of invalid request body."

        service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)"},
                                 minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"},
                                 apache={"measurement_name": "apache", "response_time_field": "mean(avg_processing_time)"})
        body = dumps(dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="sfc_1", service_functions=service_functions))
        request = testing.DummyRequest()
        request.params["from_timestamp"] = 12341412
        request.params["to_timestamp"] = 12341412
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of invalid URL parameters."

        service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)"},
                                 minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"},
                                 apache={"measurement_name": "apache", "response_time_field": "mean(avg_processing_time)"})
        body = dumps(dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="sfc_1", service_functions=service_functions))
        request = testing.DummyRequest()
        request.params["from"] = "not a timestamp"
        request.params["to"] = "not a timestamp"
        request.body = body.encode(request.charset)
        error_raised = False
        try:
            GraphAPI(request).build_temporal_graph()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "A bad request error must have been raised in case of invalid URL parameters."

        service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)"},
                                 minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"})
        build_json_body = dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="test_sfc1_1", service_functions=service_functions)
        body = dumps(build_json_body)
        request = testing.DummyRequest()
        request.params["from"] = from_timestamp
        request.params["to"] = to_timestamp
        request.body = body.encode(request.charset)
        response = GraphAPI(request).build_temporal_graph()
        graph_subresponse = response.pop("graph")
        assert response == build_json_body, "Response must contain the request body"
        assert graph_subresponse.get("uuid") is not None, "Request UUID must be attached to the response."
        request_id = graph_subresponse["uuid"]
        graph_1_id = request_id

        # check that the appropriate nodes have been created
        sf_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunction")])
        assert sf_names == {"nginx", "minio"}, "The graph must contain 2 service functions - nginx and minio"
        sf_i_names = set([node["name"] for node in graph_db.nodes.match("ServiceFunctionInstance")])
        assert sf_i_names == {"nginx_1", "minio_1"}, "The graph must contain 2 service function instances - nginx_1 and minio_1"
        endpoints = set([node["name"] for node in graph_db.nodes.match("Endpoint", uuid=request_id)])
        assert endpoints == {"minio_1_ep1", "nginx_1_ep1", "nginx_1_ep2"}

        # check the appropriate edges have been created
        self.check_exist_relationship(
            (
                ("minio_1_ep1", "Endpoint", "DC4", "ComputeNode", "hostedBy"),
                ("nginx_1_ep1", "Endpoint", "DC4", "ComputeNode", "hostedBy"),
                ("nginx_1_ep2", "Endpoint", "DC6", "ComputeNode", "hostedBy"),
                ("minio_1", "ServiceFunctionInstance", "minio_1_ep1", "Endpoint", "realisedBy"),
                ("nginx_1", "ServiceFunctionInstance", "nginx_1_ep1", "Endpoint", "realisedBy"),
                ("nginx_1", "ServiceFunctionInstance", "nginx_1_ep2", "Endpoint", "realisedBy"),
                ("minio_1", "ServiceFunctionInstance", "minio", "ServiceFunction", "instanceOf"),
                ("nginx_1", "ServiceFunctionInstance", "nginx", "ServiceFunction", "instanceOf")
            ), graph_db, request_id
        )

        # check endpoint nodes response time property
        for endpoint, response_time in (("minio_1_ep1", 9), ("nginx_1_ep1", 18.2), ("nginx_1_ep2", 22.2)):
            endpoint_node = graph_db.nodes.match("Endpoint", name=endpoint, uuid=request_id).first()
            assert endpoint_node["response_time"] == response_time, "Wrong response time property of endpoint node"

        # send a new request for a new service function chain
        service_functions = dict(minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"},
                                 apache={"measurement_name": "apache", "response_time_field": "mean(avg_processing_time)"})
        build_json_body = dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="test_sfc2_1", service_functions=service_functions)
        body = dumps(build_json_body)
        request = testing.DummyRequest()
        request.params["from"] = from_timestamp
        request.params["to"] = to_timestamp
        request.body = body.encode(request.charset)
        response = GraphAPI(request).build_temporal_graph()
        graph_subresponse = response.pop("graph")
        assert response == build_json_body, "Response must contain the request body"
        assert graph_subresponse.get("uuid") is not None, "Request UUID must be attached to the response."
        request_id = graph_subresponse["uuid"]
        graph_2_id = request_id

        assert graph_db.nodes.match("ServiceFunction", name="apache").first() is not None, "Service function apache must have been added to the graph"

        for sf_i in ("apache_1", "minio_2"):
            assert graph_db.nodes.match("ServiceFunctionInstance", name=sf_i).first() is not None, "Service function instance {0} must have been added to the graph".format(sf_i)

        for ep in ("minio_2_ep1", "apache_1_ep1"):
            assert graph_db.nodes.match("Endpoint", name=ep, uuid=request_id).first() is not None, "Endpoint {0} must have been added to the graph".format(ep)

        # check the appropriate edges have been created
        self.check_exist_relationship(
            (
                ("minio_2_ep1", "Endpoint", "DC5", "ComputeNode", "hostedBy"),
                ("apache_1_ep1", "Endpoint", "DC5", "ComputeNode", "hostedBy"),
                ("minio_2", "ServiceFunctionInstance", "minio_2_ep1", "Endpoint", "realisedBy"),
                ("apache_1", "ServiceFunctionInstance", "apache_1_ep1", "Endpoint", "realisedBy"),
                ("minio_2", "ServiceFunctionInstance", "minio", "ServiceFunction", "instanceOf"),
                ("apache_1", "ServiceFunctionInstance", "apache", "ServiceFunction", "instanceOf")
            ), graph_db, request_id
        )

        # check endpoint nodes response time property
        for endpoint, response_time in (("minio_2_ep1", 7), ("apache_1_ep1", 17.6)):
            endpoint_node = graph_db.nodes.match("Endpoint", name=endpoint, uuid=request_id).first()
            assert endpoint_node["response_time"] == response_time, "Wrong response time property of endpoint node"

    def test_delete(self, db_testing_data):
        """
        Tests the delete API endpoint of the Graph API

        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data, test database name and the graph db client
        """

        global graph_1_id, graph_2_id

        from_timestamp, to_timestamp, test_db_name, graph_db = db_testing_data

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = "invalid_graph_id"
        error_raised = False
        try:
            GraphAPI(request).delete_temporal_graph()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be raised in case of unrecognized subgraph ID"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = graph_1_id
        response = GraphAPI(request).delete_temporal_graph()
        assert response == {"uuid": graph_1_id, "deleted": 3}, "Incorrect response when deleting temporal graph"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = graph_2_id
        response = GraphAPI(request).delete_temporal_graph()
        assert response == {"uuid": graph_2_id, "deleted": 2}, "Incorrect response when deleting temporal graph"

        assert len(graph_db.nodes.match("Endpoint")) == 0, "All endpoint nodes should have been deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ComputeNode")]) == set(["DC" + str(i) for i in range(1, 7)]), "Compute nodes must not be deleted"
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunctionInstance")]) == {"nginx_1", "apache_1", "minio_1", "minio_2"}, "Service function instances must not be deleted."
        assert set([node["name"] for node in graph_db.nodes.match("ServiceFunction")]) == {"nginx", "minio", "apache"}, "Service functions must not be deleted"

    def test_rtt(self, db_testing_data):
        """
        Tests the rtt API endpoint of the Graph API

        :param db_testing_data: pair of time stamps - the from-to range of the generated influx test data, test database name and the graph db client
        """

        from_timestamp, to_timestamp, test_db_name, graph_db = db_testing_data

        service_functions = dict(nginx={"measurement_name": "nginx", "response_time_field": "mean(avg_processing_time)"},
                                 minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"})
        build_json_body = dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="test_sfc1_1", service_functions=service_functions)
        body = dumps(build_json_body)
        request = testing.DummyRequest()
        request.params["from"] = from_timestamp
        request.params["to"] = to_timestamp
        request.body = body.encode(request.charset)
        response = GraphAPI(request).build_temporal_graph()
        graph_subresponse = response.pop("graph")
        assert response == build_json_body, "Response must contain the request body"
        assert graph_subresponse.get("uuid") is not None, "Request UUID must be attached to the response."
        request_id = graph_subresponse["uuid"]

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "nginx_1_ep1"
        request.params["compute"] = "DC1"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPBadRequest:
            error_raised = True
        assert error_raised, "HTTP Bad Request must be thrown in case of missing or invalid url parameters"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "nginx_1_ep1"
        request.params["compute_node"] = "DC0"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be thrown for non existing compute node"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = "random-uuid"
        request.params["endpoint"] = "nginx_1_ep1"
        request.params["compute_node"] = "DC0"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be thrown for an endpoint node with incorrect request ID"

        request = testing.DummyRequest()
        request.matchdict["graph_id"] = request_id
        request.params["endpoint"] = "apache_1_ep1"
        request.params["compute_node"] = "DC1"
        error_raised = False
        try:
            GraphAPI(request).run_rtt_query()
        except HTTPNotFound:
            error_raised = True
        assert error_raised, "HTTP Not Found error must be thrown for a non existing endpoint"

        for dc, endpoint, forward_latencies, reverse_latencies, response_time in (
            ("DC6", "nginx_1_ep2", [], [], 22.2),
            ("DC2", "nginx_1_ep2", [22, 30, 9], [11, 26, 15], 22.2),
            ("DC3", "nginx_1_ep1", [25], [15], 18.2)
        ):
            request = testing.DummyRequest()
            request.matchdict["graph_id"] = request_id
            request.params["endpoint"] = endpoint
            request.params["compute_node"] = dc
            response = GraphAPI(request).run_rtt_query()
            assert response == {"forward_latencies": forward_latencies, "reverse_latencies": reverse_latencies, "response_time": response_time}, "Incorrect RTT response"

        # send a new request for a new service function chain
        service_functions = dict(minio={"measurement_name": "minio_http", "response_time_field": "mean(total_processing_time)/mean(total_requests_count)"},
                                 apache={"measurement_name": "apache", "response_time_field": "mean(avg_processing_time)"})
        build_json_body = dict(database=test_db_name, retention_policy="autogen", service_function_chain_instance="test_sfc2_1", service_functions=service_functions)
        body = dumps(build_json_body)
        request = testing.DummyRequest()
        request.params["from"] = from_timestamp
        request.params["to"] = to_timestamp
        request.body = body.encode(request.charset)
        response = GraphAPI(request).build_temporal_graph()
        graph_subresponse = response.pop("graph")
        assert response == build_json_body, "Response must contain the request body"
        assert graph_subresponse.get("uuid") is not None, "Request UUID must be attached to the response."
        request_id = graph_subresponse["uuid"]

        for dc, endpoint, forward_latencies, reverse_latencies, response_time in (
            ("DC5", "apache_1_ep1", [], [], 17.6),
            ("DC5", "minio_2_ep1", [], [], 7),
            ("DC3", "apache_1_ep1", [20, 30], [26, 18], 17.6),
            ("DC2", "minio_2_ep1", [22, 30], [26, 15], 7)
        ):
            request = testing.DummyRequest()
            request.matchdict["graph_id"] = request_id
            request.params["endpoint"] = endpoint
            request.params["compute_node"] = dc
            response = GraphAPI(request).run_rtt_query()
            assert response == {"forward_latencies": forward_latencies, "reverse_latencies": reverse_latencies, "response_time": response_time}, "Incorrect RTT response"

    @staticmethod
    def check_exist_relationship(relationships_tuple, graph, uuid):
        for relationship in relationships_tuple:
            from_node_name, from_node_type, to_node_name, to_node_type, relationship_type = relationship
            if from_node_type == "Endpoint":
                from_node = graph.nodes.match(from_node_type, name=from_node_name, uuid=uuid).first()
            else:
                from_node = graph.nodes.match(from_node_type, name=from_node_name).first()
            assert from_node is not None

            if to_node_type == "Endpoint":
                to_node = graph.nodes.match(to_node_type, name=to_node_name, uuid=uuid).first()
            else:
                to_node = graph.nodes.match(to_node_type, name=to_node_name).first()
            assert to_node is not None

            assert graph.relationships.match(nodes=(from_node, to_node), r_type=relationship_type).first() is not None, "Graph is missing a required relationship"
